#include "Datetime.h"

DateTime::DateTime() {}
DateTime::DateTime(int y, int m, int d) {
	_year = y;
	_month = m;
	_day = d;
	updatetp();
}
DateTime::DateTime(Timepoint tp) {
	set_timepoint(tp);
}

int DateTime::Year() {
	return _year;
}
int DateTime::Year(int y) {
	_year = y;
	updatetp();
	return _year;
}
int DateTime::Month() {
	return _month;
}
int DateTime::Month(int m) {
	_month = m;
	updatetp();
	return _month;
}
int DateTime::Day() {
	return _day;
}
int DateTime::Day(int d) {
	_day = d;
	updatetp();
	return _day;
}
double DateTime::TotalDays() {
	return _tp.Days();
}
int DateTime::Hour() {
	return _hour;
}
int DateTime::Hour(int h) {
	_hour = h;
	updatetp();
	return _hour;
}
double DateTime::TotalHours() {
	return _tp.Hours();
}
int DateTime::Minute() {
	return _minute;
}
int DateTime::Minute(int m) {
	_minute = m;
	updatetp();
	return _minute;
}
double DateTime::TotalMinutes() {
	return _tp.Minutes();
}
int DateTime::Second() {
	return _second;
}
int DateTime::Second(int s) {
	_second = s;
	updatetp();
	return _second;
}
double DateTime::TotalSeconds() {
	return _tp.Seconds();
}
double DateTime::Millisecond() {
	return _nanos / 1e6;
}
double DateTime::TotalMilliseconds() {
	return _tp.Milliseconds();
}
double DateTime::Microsecond() {
	return _nanos / 1e3;
}
double DateTime::TotalMicroseconds() {
	return _tp.Microseconds();
}
double DateTime::Nanosecond() {
	return _nanos;
}
double DateTime::Nanosecond(double ns) {
	_nanos = ns;
	updatetp();
	return _nanos;
}
double DateTime::TotalNanoseconds() {
	return _tp.Nanoseconds();
}
void DateTime::updatetp() {
	double n = 0;
	n += _nanos;
	n += 1e9 * _second;
	n += 6e10 * _minute;
	n += 3.6e12 * _hour;
	n += 8.64e13 * (double)days_from_civil(_year, _month, _day);
}
void DateTime::set_timepoint(Timepoint tp) {
	DateTime dt0 = civil_from_days(tp.Days());
	_year = dt0._year;
	_month = dt0._month;
	_day = dt0._day;
	_hour = (int)tp.Hours() - 24 * (int)tp.Days();
	_minute = (int)tp.Minutes() - 60 * (int)tp.Hours();
	_second = (int)tp.Seconds() - 60 * (int)tp.Minutes();
	_nanos = tp.Nanoseconds() - 1e9 * (int)tp.Seconds();
}
DateTime::operator Timepoint() {
	return _tp;
}

//Public domain conversion functions
int DateTime::days_from_civil(int y, unsigned m, unsigned d) //noexcept
{
	y -= m <= 2;
	const int era = (y >= 0 ? y : y - 399) / 400;
	const unsigned yoe = static_cast<unsigned>(y - era * 400);      // [0, 399]
	const unsigned doy = (153 * (m + (m > 2 ? -3 : 9)) + 2) / 5 + d - 1;  // [0, 365]
	const unsigned doe = yoe * 365 + yoe / 4 - yoe / 100 + doy;         // [0, 146096]
	return era * 146097 + static_cast<int>(doe) - 719468;
}
DateTime DateTime::civil_from_days(int z) //noexcept
{
	z += 719468;
	const int era = (z >= 0 ? z : z - 146096) / 146097;
	const unsigned doe = static_cast<unsigned>(z - era * 146097);          // [0, 146096]
	const unsigned yoe = (doe - doe / 1460 + doe / 36524 - doe / 146096) / 365;  // [0, 399]
	const int y = static_cast<int>(yoe) + era * 400;
	const unsigned doy = doe - (365 * yoe + yoe / 4 - yoe / 100);                // [0, 365]
	const unsigned mp = (5 * doy + 2) / 153;                                   // [0, 11]
	const unsigned d = doy - (153 * mp + 2) / 5 + 1;                             // [1, 31]
	const unsigned m = mp + (mp < 10 ? 3 : -9);                            // [1, 12]
	return DateTime(y + (m <= 2), m, d);
}
Weekday DateTime::weekday_from_days(int z) //noexcept
{
	return static_cast<Weekday>(z >= -4 ? (z + 4) % 7 : (z + 5) % 7 + 6);
}