#pragma once

#ifndef TIMEKEEPER_H
#define TIMEKEEPER_H

#include "Timepoint.h"
#include "Duration.h"

///	Useful for keeping track of time data throughout a program.
///	When using discrete mode, you must have a class named Manager
///	that periodically calls apply_timestep;
///	otherwise, all data values will remain the same
class TimeKeeper {
	friend class Manager;
public:
	enum Mode {
		Continuous = 0,
		Discrete = 1,
	};

	TimeKeeper();
	TimeKeeper(int mode);
	static void HangUntil(Timepoint);
	template <class T>
	static void HangUntil(bool(*_test)(T), T);
	static void HangUntil(bool(*_test)(void));
	static void HangFor(Duration);
	Timepoint Now();
	double Now(TimeUnit);
	double DeltaTime();
	Duration Runtime();
	double StepsPerSecond();
private:
	Mode _mode;

	Timepoint _start;
	Timepoint _stepstart;
	Timepoint _prevstepstart;

	enum Dirtyflags : unsigned int {
		deltat = 1,
		sps = 2,
	};
	int _dirty;		//	dirty flags
	double _deltat;
	double _sps;

	//	Operator methods
	void apply_timestep();
};


#endif	//	TIMEKEEPER_H